import { Address, AddressRequestBody, Cart, CheckoutRequestBody, CheckoutSelectors, Consignment, ConsignmentAssignmentRequestBody, Country, Customer, CustomerRequestOptions, FormField, ShippingInitializeOptions, ShippingRequestOptions } from '@bigcommerce/checkout-sdk'
import { noop } from 'lodash'
import React, { Component, ReactNode } from 'react'
import { connect } from 'react-redux'
import { createSelector } from 'reselect'

import { shouldUseStripeLinkByMinimumAmount } from '@bigcommerce/checkout/instrument-utils';
import { CheckoutContextProps } from '@bigcommerce/checkout/payment-integration-api';
import { AddressFormSkeleton } from '@bigcommerce/checkout/ui';

import { isEqualAddress, mapAddressFromFormValues } from '../address';
import B3Checkout, {  NoAddressTip } from '../b3checkout'
import store from '../b3checkout/store'
import { setGlobalIsPending, setGlobalLoadingStage } from '../b3checkout/store/actions'
import { withCheckout } from '../checkout';
import CheckoutStepStatus from '../checkout/CheckoutStepStatus';
import { EMPTY_ARRAY, isFloatingLabelEnabled } from '../common/utility';
import getProviderWithCustomCheckout from '../payment/getProviderWithCustomCheckout';
import { PaymentMethodId } from '../payment/paymentMethod';

import { UnassignItemError } from './errors'
import getShippableItemsCount from './getShippableItemsCount'
import getShippingMethodId from './getShippingMethodId'
import { MultiShippingFormValues } from './MultiShippingForm'
import ShippingForm from './ShippingForm'
import ShippingHeader from './ShippingHeader'
import { SingleShippingFormValues } from './SingleShippingForm'
import StripeShipping from './stripeUPE/StripeShipping';
import Swal from 'sweetalert2'


export interface ShippingProps {
    isBillingSameAsShipping: boolean;
    cartHasChanged: boolean;
    isMultiShippingMode: boolean;
    step: CheckoutStepStatus;
    onCreateAccount(): void;
    onToggleMultiShipping(): void;
    onReady?(): void;
    onUnhandledError(error: Error): void;
    onSignIn(): void;
    navigateNextStep(isBillingSameAsShipping: boolean): void;
}

export interface WithCheckoutShippingProps {
    billingAddress?: Address
    cart: Cart
    consignments: Consignment[]
    countries: Country[]
    countriesWithAutocomplete: string[]
    customer: Customer
    customerMessage: string
    googleMapsApiKey: string
    isGuest: boolean
    isInitializing: boolean
    isLoading: boolean
    isShippingStepPending: boolean
    methodId?: string
    shippingAddress?: Address
    shouldShowAddAddressInCheckout: boolean;
    shouldShowMultiShipping: boolean;
    shouldShowOrderComments: boolean;
    providerWithCustomCheckout?: string;
    isFloatingLabelEnabled?: boolean;
    assignItem(consignment: ConsignmentAssignmentRequestBody): Promise<CheckoutSelectors>;
    deinitializeShippingMethod(options: ShippingRequestOptions): Promise<CheckoutSelectors>;
    deleteConsignments(): Promise<Address | undefined>;
    getFields(countryCode?: string): FormField[];
    initializeShippingMethod(options: ShippingInitializeOptions): Promise<CheckoutSelectors>;
    loadShippingAddressFields(): Promise<CheckoutSelectors>;
    loadBillingAddressFields(): Promise<CheckoutSelectors>;
    loadShippingOptions(): Promise<CheckoutSelectors>;
    signOut(options?: CustomerRequestOptions): void;
    createCustomerAddress(address: AddressRequestBody): Promise<CheckoutSelectors>;
    unassignItem(consignment: ConsignmentAssignmentRequestBody): Promise<CheckoutSelectors>;
    updateBillingAddress(address: Partial<Address>): Promise<CheckoutSelectors>;
    updateCheckout(payload: CheckoutRequestBody): Promise<CheckoutSelectors>;
    updateShippingAddress(address: Partial<Address>): Promise<CheckoutSelectors>;
    shouldRenderStripeForm: boolean;
}

interface ShippingState {
    isInitializing: boolean
}

interface B3Props {
    b3Company: any
    isB3User: boolean
    checkoutSettings: any
}

class Shipping extends Component<B3Props & ShippingProps & WithCheckoutShippingProps, ShippingState> {
    constructor(props: B3Props & ShippingProps & WithCheckoutShippingProps) {
        super(props)

        this.state = {
            isInitializing: true,
        }
    }

    async componentDidMount(): Promise<void> {
        const {
            loadShippingAddressFields,
            loadBillingAddressFields,
            loadShippingOptions,
            onReady = noop,
            onUnhandledError = noop,
        } = this.props

        try {
            store.dispatch(setGlobalIsPending(true))
            store.dispatch(setGlobalLoadingStage('common.loading_stages.address'))
            await Promise.all([loadShippingAddressFields(), loadShippingOptions(), loadBillingAddressFields()]);

            onReady()
        } catch (error) {
            onUnhandledError(error)
        } finally {
            this.setState({ isInitializing: false })
            store.dispatch(setGlobalIsPending(false))
        }
    }

    get shouldShowSaveAddress() {
        const {
            isGuest,
            isB3User,
        } = this.props

        const {
            storage,
            constants,
        } = B3Checkout.utils

        const B3RoleId = storage.B3RoleId.getValue()

        const { ADMIN, SENIOR } = constants.B3Role
     
        return isB3User
            ? [ADMIN, SENIOR].includes(B3RoleId)
            : !isGuest
    }

    get shouldShowSaveAddressInCheckoutSettings() {
        const {
            checkoutSettings,
        } = this.props

        return checkoutSettings['shipping.address.save']
    }

    get shouldAllowNewAddress() {
        const { checkoutSettings } = this.props

        return checkoutSettings['shipping.address.add']
    }

    get addresses() {
        const {
            customer,
            isB3User,
            b3Company,
        } = this.props

        return isB3User
            ? b3Company.addresses.filter((address: { isShipping: number }) => address.isShipping === 1)
            : customer.addresses
    }

    render(): ReactNode {
        const {
            isBillingSameAsShipping,
            isGuest,
            shouldShowMultiShipping,
            customer,
            isB3User,
            updateShippingAddress,
            initializeShippingMethod,
            deinitializeShippingMethod,
            isMultiShippingMode,
            step,
            isFloatingLabelEnabled,
            shouldRenderStripeForm,
            ...shippingFormProps
        } = this.props

        const {
            isInitializing,
        } = this.state

        if (shouldRenderStripeForm && !customer.email && this.props.countries.length > 0) {
            return <StripeShipping
                { ...shippingFormProps }
                customer={ customer }
                deinitialize={deinitializeShippingMethod}
                initialize={initializeShippingMethod}
                isBillingSameAsShipping={isBillingSameAsShipping}
                isGuest={ isGuest }
                isLoading={ isInitializing }
                isMultiShippingMode={isMultiShippingMode}
                isShippingMethodLoading={ this.props.isLoading }
                onMultiShippingChange={ this.handleMultiShippingModeSwitch }
                onSubmit={this.handleSingleShippingSubmit}
                shouldShowMultiShipping={ shouldShowMultiShipping }
                step={step}
                updateAddress={updateShippingAddress}
            />;
        }

        const isShouldShowNewAddressForm = !(isB3User ? !this.addresses.length && !this.shouldAllowNewAddress : false)

        return (
            <AddressFormSkeleton isLoading={isInitializing}>
                <div className="checkout-form">
                    {
                        isShouldShowNewAddressForm ? (
                            <>
                                <ShippingHeader
                                    isGuest={isGuest}
                                    isMultiShippingMode={isMultiShippingMode}
                                    onMultiShippingChange={this.handleMultiShippingModeSwitch}
                                    shouldShowMultiShipping={shouldShowMultiShipping}
                                />
                                <ShippingForm
                                    {...shippingFormProps}
                                    addresses={this.addresses}
                                    deinitialize={deinitializeShippingMethod}
                                    initialize={initializeShippingMethod}
                                    isBillingSameAsShipping={isBillingSameAsShipping}
                                    isFloatingLabelEnabled={isFloatingLabelEnabled}
                                    isGuest={isGuest}
                                    isMultiShippingMode={isMultiShippingMode}
                                    onMultiShippingSubmit={this.handleMultiShippingSubmit}
                                    onSingleShippingSubmit={this.handleSingleShippingSubmit}
                                    onUseNewAddress={this.handleUseNewAddress}

                                    shouldShowSaveAddress={ this.shouldShowSaveAddress && this.shouldShowSaveAddressInCheckoutSettings }
                                    updateAddress={updateShippingAddress}
                                />
                            </>
                        ) : (<NoAddressTip type="shipping" />)
                    }
                    
                </div>
            </AddressFormSkeleton>
        );
    }

    private handleMultiShippingModeSwitch: () => void = async () => {
        const {
            consignments,
            isMultiShippingMode,
            onToggleMultiShipping = noop,
            onUnhandledError = noop,
            updateShippingAddress,
        } = this.props;

        if (isMultiShippingMode && consignments.length > 1) {
            this.setState({ isInitializing: true });

            try {
                // Collapse all consignments into one
                await updateShippingAddress(consignments[0].shippingAddress);
            } catch (error) {
                onUnhandledError(error);
            } finally {
                this.setState({ isInitializing: false });
            }
        }

        onToggleMultiShipping();
    };

    private handleSingleShippingSubmit: (values: SingleShippingFormValues) => void = async ({
        billingSameAsShipping,
        shippingAddress: addressValues,
        b2bCustomFields,
        orderComment,
    }) => {
        const {
            cart,
            customer,
            customerMessage,
            isGuest,
            updateCheckout,
            updateShippingAddress,
            updateBillingAddress,
            navigateNextStep,
            onUnhandledError,
            shippingAddress,
            billingAddress,
            methodId,
            isB3User,
            checkoutSettings,
        } = this.props

        const updatedShippingAddress = addressValues && mapAddressFromFormValues(addressValues)
        const promises: Array<Promise<CheckoutSelectors>> = []
        const hasRemoteBilling = this.hasRemoteBilling(methodId)

        if (isB3User && b2bCustomFields) {
            B3Checkout.extrasFiles.saveExtrasAddress(b2bCustomFields, 'shipping')
        }

        if (isB3User && checkoutSettings['shipping.address.save']) {
            await B3Checkout.addresses.addNewB3Address({
                newAddress: updatedShippingAddress,
                billingSameAsShipping,
                isBillingMode: false,
                checkoutSettings,
                b2bCustomFields: B3Checkout.extrasFiles.decodeAddressExtraFieldsKeys(b2bCustomFields),
            })
        }

        if (isB3User && billingSameAsShipping) {
            B3Checkout.utils.storage.selectBillAddressId.setValue(B3Checkout.utils.storage.selectShoppingAddressId.getValue() || '')
        }

        if (!isEqualAddress(updatedShippingAddress, shippingAddress) || shippingAddress?.shouldSaveAddress !== updatedShippingAddress?.shouldSaveAddress) {
            promises.push(updateShippingAddress(updatedShippingAddress || {}));
        }

        if (
            billingSameAsShipping &&
            updatedShippingAddress &&
            !isEqualAddress(updatedShippingAddress, billingAddress) &&
            !hasRemoteBilling
        ) {
            if (
                !isB3User ||
                (
                    isB3User
                    && checkoutSettings['shipping.singleAddress.allowAsBilling']
                )
            ) {
                if (isB3User) B3Checkout.extrasFiles.billingSameAsShippingAddress()
                promises.push(updateBillingAddress(updatedShippingAddress))
            }
        }

        if (customerMessage !== orderComment) {
            promises.push(updateCheckout({ customerMessage: orderComment }))
        }

        const restrictionsRules = (window as any).product_restrictions || []
        const addressState = shippingAddress?.stateOrProvinceCode;
        const lineItems = cart.lineItems;

        for (let rule of restrictionsRules) {
            const states = rule.allowState || [];
            const emails = rule.exceptionCustomer || [];
            const groups = rule.exceptionGroup || [];

            let shouldRemove = false;
            if (addressState && !states.includes(addressState) && !emails.includes(customer.email)) {
                shouldRemove = true;
                if (!isGuest && customer.customerGroup && groups.includes(customer.customerGroup.name)) {
                    shouldRemove = false;
                }
            }
            if (shouldRemove) {
                let removeItemNames: Array<string | number> = [];
                let removeItemIds: Array<string | number> = [];
                for (let item of lineItems.physicalItems) {
                    // let parentSku = item.sku;
                    // if (item!.options!.length > 0) {
                    //     parentSku = await this.getSkuById(item.productId);
                    // }
                    if ('sku' in rule && (rule.sku == item.sku || rule.sku == item.sku) || 'brand' in rule && rule.brand == item.brand) {
                        removeItemNames.push(item.name)
                        removeItemIds.push(item.id)
                    }
                }
                for (let item of lineItems.digitalItems) {
                    // let parentSku = item.sku;
                    // if (item!.options!.length > 0) {
                    //     parentSku = await this.getSkuById(item.productId);
                    // }
                    if ('sku' in rule && (rule.sku == item.sku || rule.sku == item.sku) || 'brand' in rule && rule.brand == item.brand) {
                        removeItemNames.push(item.name)
                        removeItemIds.push(item.id)
                    }
                }
                if (removeItemNames.length > 0) {
                    await Swal.fire({
                        icon: 'warning',
                        html: '<p>One or more of the items in your cart cannot be purchased in your location due to vendor restrictions.</p>' +
                            '<p>Following items will be removed, you can review your cart and continue shopping:</p>' + 
                            '<p>' + removeItemNames.join(', ') + '</p>',
                        confirmButtonText: 'OK',
                        allowOutsideClick: false,
                        allowEscapeKey: false,
                    }).then(async (result) => {
                        if (result.isConfirmed) {
                            let deleteCartItems = [];
                            for (let itemId of removeItemIds) {
                                deleteCartItems.push(
                                    fetch('/api/storefront/carts/' + cart.id + '/items/' + itemId, {
                                        method: "DELETE",
                                        credentials: "same-origin",
                                        headers: {
                                          "Content-Type": "application/json",
                                        }
                                    })
                                    .then(response => response.json())
                                    .catch(error => console.error(error))
                                )
                            }
                            await Promise.all(deleteCartItems)

                            await Swal.fire({
                                icon: 'success',
                                html: '<p>Items removed:</p>' + '<p>' + removeItemNames.join(', ') + '</p>',
                                confirmButtonText: 'Back to Cart',
                                allowOutsideClick: false,
                                allowEscapeKey: false,
                            });
                            window.location.replace('/cart.php')
                        }
                    })
                }
            }
        }

        try {
            await Promise.all(promises)

            navigateNextStep(billingSameAsShipping)
        } catch (error) {
            if (error instanceof Error) {
                onUnhandledError(error);
            }
        }
    };

    private hasRemoteBilling: (methodId?: string) => boolean = (methodId) => {
        const PAYMENT_METHOD_VALID = ['amazonpay'];

        return PAYMENT_METHOD_VALID.some((method) => method === methodId);
    };

    private handleUseNewAddress: (address: Address, itemId: string) => void = async (
        address,
        itemId,
    ) => {
        const { unassignItem, onUnhandledError } = this.props;

        try {
            await unassignItem({
                address,
                lineItems: [
                    {
                        quantity: 1,
                        itemId,
                    },
                ],
            });

            location.href = '/account.php?action=add_shipping_address&from=checkout';
        } catch (error) {
            if (error instanceof UnassignItemError) {
                onUnhandledError(new UnassignItemError(error));
            }
        }
    };

    private handleMultiShippingSubmit: (values: MultiShippingFormValues) => void = async ({
        orderComment,
    }) => {
        const { customerMessage, updateCheckout, navigateNextStep, onUnhandledError } = this.props;

        try {
            if (customerMessage !== orderComment) {
                await updateCheckout({ customerMessage: orderComment })
            }

            navigateNextStep(false)
        } catch (error) {
            if (error instanceof Error) {
                onUnhandledError(error);
            }
        }
    };
}

const deleteConsignmentsSelector = createSelector(
    ({ checkoutService: { deleteConsignment } }: CheckoutContextProps) => deleteConsignment,
    ({ checkoutState: { data } }: CheckoutContextProps) => data.getConsignments(),
    (deleteConsignment, consignments) => async () => {
        if (!consignments || !consignments.length) {
            return
        }

        const [{ data }] = await Promise.all(consignments.map(({ id }) => deleteConsignment(id)));

        return data.getShippingAddress();
    },
);

// tslint:disable-next-line:cyclomatic-complexity
export function mapToShippingProps({
    checkoutService,
    checkoutState,
}: CheckoutContextProps): WithCheckoutShippingProps | null {
    const {
        data: {
            getCart,
            getCheckout,
            getConfig,
            getCustomer,
            getConsignments,
            getShippingAddress,
            getBillingAddress,
            getShippingAddressFields,
            getShippingCountries,
        },
        statuses: {
            isShippingStepPending,
            isSelectingShippingOption,
            isLoadingShippingOptions,
            isUpdatingConsignment,
            isCreatingConsignments,
            isCreatingCustomerAddress,
            isLoadingShippingCountries,
            isUpdatingBillingAddress,
            isUpdatingCheckout,
        },
    } = checkoutState

    const checkout = getCheckout()
    const config = getConfig()
    const consignments = getConsignments() || []
    const customer = getCustomer()
    const cart = getCart()

    if (!checkout || !config || !customer || !cart) {
        return null
    }

    const {
        checkoutSettings: {
            enableOrderComments,
            features,
            hasMultiShippingEnabled,
            googleMapsApiKey,
        },
    } = config

    const methodId = getShippingMethodId(checkout, config);
    const shippableItemsCount = getShippableItemsCount(cart);
    const isLoading =
        isLoadingShippingOptions() ||
        isSelectingShippingOption() ||
        isUpdatingConsignment() ||
        isCreatingConsignments() ||
        isUpdatingBillingAddress() ||
        isUpdatingCheckout() ||
        isCreatingCustomerAddress();
    const shouldShowMultiShipping =
        hasMultiShippingEnabled && !methodId && shippableItemsCount > 1;
    const countriesWithAutocomplete = ['US', 'CA', 'AU', 'NZ'];

    if (features['CHECKOUT-4183.checkout_google_address_autocomplete_uk']) {
        countriesWithAutocomplete.push('GB')
    }

    const shippingAddress =
        !shouldShowMultiShipping && consignments.length > 1 ? undefined : getShippingAddress();

    const providerWithCustomCheckout = getProviderWithCustomCheckout(
        config.checkoutSettings.providerWithCustomCheckout,
    );

    return {
        assignItem: checkoutService.assignItemsToAddress,
        billingAddress: getBillingAddress(),
        cart,
        consignments,
        countries: getShippingCountries() || EMPTY_ARRAY,
        countriesWithAutocomplete,
        customer,
        customerMessage: checkout.customerMessage,
        createCustomerAddress: checkoutService.createCustomerAddress,
        deinitializeShippingMethod: checkoutService.deinitializeShipping,
        deleteConsignments: deleteConsignmentsSelector({ checkoutService, checkoutState }),
        getFields: getShippingAddressFields,
        googleMapsApiKey,
        initializeShippingMethod: checkoutService.initializeShipping,
        isGuest: customer.isGuest,
        isInitializing: isLoadingShippingCountries() || isLoadingShippingOptions(),
        isLoading,
        isShippingStepPending: isShippingStepPending(),
        loadShippingAddressFields: checkoutService.loadShippingAddressFields,
        loadBillingAddressFields: checkoutService.loadBillingAddressFields,
        loadShippingOptions: checkoutService.loadShippingOptions,
        methodId,
        providerWithCustomCheckout,
        shippingAddress,
        shouldShowMultiShipping,
        shouldShowAddAddressInCheckout:
            features['CHECKOUT-4726.add_address_in_multishipping_checkout'],
        shouldShowOrderComments: enableOrderComments,
        signOut: checkoutService.signOutCustomer,
        unassignItem: checkoutService.unassignItemsToAddress,
        updateBillingAddress: checkoutService.updateBillingAddress,
        updateCheckout: checkoutService.updateCheckout,
        updateShippingAddress: checkoutService.updateShippingAddress,
        isFloatingLabelEnabled: isFloatingLabelEnabled(config.checkoutSettings),
        shouldRenderStripeForm: providerWithCustomCheckout === PaymentMethodId.StripeUPE && shouldUseStripeLinkByMinimumAmount(cart),
    };
}

const mapStateToProps = (state: { global: { isB3User: boolean, checkoutSettings: any }, company: any }) => {
    const {
        global: {
            isB3User,
            checkoutSettings,
        },
        company: b3Company
    } = state

    return {
        b3Company,
        isB3User,
        checkoutSettings,
    }
}

export default connect(mapStateToProps, {})(withCheckout(mapToShippingProps)(Shipping))
